/* pfs.h
 * 1997/10/24
 * Jeff Weeks
 *
 * This is the main header file (object) for the Poly One File system.
 * Currently is has functions to write blocks, and format a drive only.
 */

#include <dpmi.h>
#include <string.h>

#ifndef __poly_one_fs__
#define __poly_one_fs__

// The following defines are just for convenience
#define  INODE_SIZE     128

// The following defines are for file types
#define  SPECIAL        0

// The following defines are for subtypes (of type special, anyway)
#define  INODE_TABLE    0
#define  INODE_BITMAP   1
#define  BLOCK_BITMAP   2
#define  DIRECTORY      'D'   // A directory file
#define  LINK           'L'   // A file link
#define  OS             'K'   // the operating system binary (kernel)

// some typedefs for fluid reading and writting of this code :)
// I may switch to the byte, word, dword system soon.  I don't know.
typedef unsigned long  u32;
typedef signed   long  s32;
typedef unsigned short u16;
typedef signed   short s16;
typedef unsigned char  u8;
typedef signed   char  s8;

// The format of the poly one super block.  No need to pack the structure, it's
// already dword aligned and optomized with padding.
typedef struct SuperBlock {
  u32 magic;
  u32 os_inode;
  u32 root_inode;
  u32 inode_table;
  u32 inode_bitmap;
  u32 block_bitmap;
  u8  block_size;
  u8  state;
  u8  error_behaviour;
  u8  pad;
  u16 os_version;
  u16 fs_version;
  u32 creator_os;
  u32 num_inodes;
  u32 free_inodes;
  u32 num_blocks;
  u32 free_blocks;
  u32 first_data_block;                // This will probably disapear soon
  u16 max_mnt_count;
  u16 mnt_count;
  u8  pad_to_1k[1024 - 60];
};

// The format of an Inode.  Again, dword aligned and optomized.
typedef struct Inode {
  u16 type;
  u16 subtype;
  u16 permissions;
  u16 links_count;
  u32 size;
  u32 access_time;
  u32 creation_time;
  u32 modification_time;
  u32 user_id;
  u32 group_id;
  u32 flags;
  u32 reserved1;
  u32 reserved2;
  u32 reserved3;
  u32 start_block_1;
  u32 length_1;
  u32 start_block_2;
  u32 length_2;
  u32 start_block_3;
  u32 length_3;
  u32 start_block_4;
  u32 length_4;
  u32 start_block_5;
  u32 length_5;
  u32 start_block_6;
  u32 length_6;
  u32 start_block_7;
  u32 length_7;
  u32 start_block_8;
  u32 length_8;
  u32 start_block_9;
  u32 length_9;
  u32 start_block_10;
  u32 length_10;
};

// The structure of a directory.  Variable sized, but still optomized.
typedef struct Directory {
  u32 file_inode;
  u32 attrib_inode;
  u16 record_length;
  u8  name_length;
  u8  character_encoding;
  u8  name[256];
};

// Finally, the main PFS class
class PFS {
  public:
    SuperBlock super;                  // a superblock structure
    Inode inode;                       // an inode structure
    Directory dir;                     // a directory structure
    __dpmi_regs regs;                  // just incase it's needed
    u8 drive;                          // the current drive number
    char block_buffer[4096];           // a cached block buffer
    u32 block_size;                    // for convenience we store the block size
    u16 cylinders, sectors, heads;     // some drive info variables
    u16 sectors_per_block;             // number of sectors in one block
    u32 size;                          // the size of the whole disk

    PFS(void);                         // just setup some defaults

    void set_drive(u8 num);
    void format(int _block_size = 1024, int _num_inodes = 64);
    void write_blocks(char *buf, int start, int total);
};

PFS::PFS(void) {
  drive = 0;                           // by default, use the floppy drive!
  block_size = 0;
  cylinders = sectors = heads = 0;
  sectors_per_block = 0;
}

void PFS::set_drive(u8 num) {
  // First, we'll get some info about this drive.  We use the block_buffer
  // as a transfer buffer, which is actually bigger than needed, however,
  // it's convenient for the moment :)
  if(biosdisk(0x08, num, 0, 0, 0, 0, &block_buffer)) {
    cout << "Drive not valid: " << (int)num;
    exit(1);
  }

  // setup up our disk variables
  drive     = num;
  cylinders = (int) ((int)(block_buffer[0] & 0xC0) << 2) + block_buffer[1];
  sectors   = block_buffer[0] & 0x3F;
  heads     = block_buffer[3];
  size      = (cylinders+1) * sectors * (heads+1) * 512;

  // spit out some info
  cout << "Using drive: " << (int)drive
       << "\t" << "Logical disk size: " << size << endl;
  cout << cylinders+1 << " cylinders\n";
  cout << sectors << " sectors\n";
  cout << heads+1 << " heads\n";
}

/* This is a fairly difficult fuction to, not only understand, but write!
 * For both of these reasons I'll include a juicy description of what's
 * going on.  First of all, what we want is to create the superblock, inode
 * table, inode bitmap, and the block bitmap.  It seems logical to do all
 * these things in one function because they are all inter-related.  Change
 * one thing, and the others are affected.  The root directory probably
 * should be created here as well.
 *
 * We also do as much as possible in memory.  Infact, all structures are
 * created in memory first then written in one pass.  This is both fast,
 * and less confusing.
 */
void PFS::format(int _block_size = 1024, int _num_inodes = 64) {
  char *ibitmap, *bbitmap;  // inode and block bitmap buffers
  char *itable;             // the inode table
  int _num_blocks;
  int alloc_blocks = 0;     // contains a running total of allocated blocks
  int alloc_inodes = 0;     // and inodes.  They'll be flagged later.
  int i;
  int inode_block,          // each of these variables hold the
      block_block,          // block location of certain vital
      root_block;           // parts of the Poly One FS.
  int inode_length,         // ... and each of these variables hold
      block_length,         // the length of the above parts
      table_length;

  // Setup some variables for the PFS class
  block_size = _block_size;
  sectors_per_block = block_size / 512;

  // First we'll create a superblock in memory
  cout << "\nCreating superblock...\n";
  cout << "  Using block size of " << _block_size << endl;
  cout << "  Creating " << _num_inodes << " inodes\n";
  cout << "  " << (int)(size / _block_size) << " blocks\n";
  cout << "  " << (int)(size / _num_inodes) << " inode spacing\n";
  cout << "Creating Superblock with the following specs:\n"
       << "  Created for PolyOS 1.0\n"
       << "  FS Version 1.0, Poly One\n";

  // Setup the superblock in memory
  super.magic = (int)('P' | 'F' << 8 | 'S' << 16 | '1' << 24); // will say PFS1
  super.os_inode = 0;      // inode zero should be considered non-bootable (no OS)
  super.root_inode = 3;    // root directory will be inode 3
  // Where should we put the inode table?  We cannot corrupt superblock!
  // Note, the location of the itable is a block location, not an inode!
  if(_block_size == 1024) super.inode_table = 2;
  else super.inode_table = 1;
  super.inode_bitmap = 1;  // inode which describes location of inode bitmap
  super.block_bitmap = 2;  // inode which describes location of block bitmap
  super.block_size = _block_size >> 11;
  super.state = super.error_behaviour = 0;
  super.os_version = super.fs_version = 256; // version 1.0
  super.creator_os = 0;    // PolyOS
  super.num_inodes = super.free_inodes = _num_inodes;
  super.num_blocks = super.free_blocks = (size / _block_size);
  super.first_data_block = 0; // Is this really even needed?
  super.max_mnt_count = 500;
  super.mnt_count = 0;

  // _num_blocks is used later on, so lets set it up correctly
  _num_blocks = super.num_blocks;

  // We'll have to set the first block or two to allocated so we don't
  // overwrite the superblock or bootsector.  This can be done simply
  // by adding super.inode_table to our allocated blocks total.
  alloc_blocks = super.inode_table;

  // Here's one of the easier things to do;  Create the inode table in mem.
  // However, we'll have to round up the nearest block.  Atleast we do now.
  // Later on, I'll only allow certain inode numbers which will fill a block
  // totally.  That'll have to wait 'till the GUI version of pformat though.
  cout << "Creating inode table\n";
  i = (super.num_inodes * INODE_SIZE);

  // Find out how many blocks are needed to allocate the inode table
  // and round up if necessary.
  if( (i & (block_size-1)) == 0) i /= block_size;
  else i = i / block_size + 1;

  // Allocate room for the inode table, and flag the blocks as allocated
  itable = new char[i * block_size];
  alloc_blocks += i;

  // Lets zero it out too, since it's easier to read on the disk when
  // debugging ;)
  memset(itable, 0, i * block_size);

  // Now we must create the inode and block bitmaps in memory.  The problem
  // is that we don't know how big they are.  We can calculate this out as
  // follows, though...

  // Find out how many bytes are in each bitmap (remember to round up if
  // nescessary)
  cout << "Calculating inode and block bitmap sizes\n";
  if( (_num_inodes & 7) == 0) _num_inodes >>= 3;
  else _num_inodes = (_num_inodes >> 3) + 1;

  if( (_num_blocks & 7) == 0) _num_blocks >>= 3;
  else _num_blocks = (_num_blocks >> 3) + 1;

  // Now find out how many blocks are needed to store each bitmap (again,
  // round up if nescessary)
  if( (_num_inodes & (block_size-1)) == 0) _num_inodes /= block_size;
  else _num_inodes = _num_inodes / block_size + 1;

  if( (_num_blocks & (block_size-1)) == 0) _num_blocks /= block_size;
  else _num_blocks = _num_blocks / block_size + 1;

  // Again, it should be noted that, ideally, the number of inodes and
  // blocks SHOULD be constrained to fill the whole block in the
  // block and inode bitmap and the inode table.  The great thing about
  // having 128 byte inodes is that the above is completely possible.

  // Okay, now _num_inodes and _num_blocks contain the number of blocks
  // required to store both the inode bitmap and the block bitmap so
  // we can allocate room for the bitmaps and zero them out.  I suppose
  // I should rename those variables... they no longer reflect what
  // they hold :)
  cout << "Allocating and clearing inode and block bitmaps\n";
  ibitmap = new char[_num_inodes * block_size];
  bbitmap = new char[_num_blocks * block_size];
  memset(ibitmap,0, _num_inodes * block_size );
  memset(bbitmap,0, _num_blocks * block_size );
  alloc_blocks += _num_inodes;
  alloc_blocks += _num_blocks;

  // Create the first inode in the inode table.  This inode will reference
  // the inode table itself.  Remember to add one to our allocated inode
  // running total.  What's important to note here is that inode 0 will
  // _always_ reference the inode table.  Also, in the superblock, inode
  // 0 is referenced by it's block number, not it's inode number (because,
  // afterall, an inode number does us no good when we don't know where the
  // inode table is!)
  cout << "Creating special inodes\n";
  // Inode 0 = inode table (ALWAYS!)
  inode.type = SPECIAL;
  inode.subtype = INODE_TABLE;
  inode.size = super.num_inodes * INODE_SIZE;
  // The permissions, etc, are irrelavent here.
  inode.start_block_1 = super.inode_table;
  inode.length_1 = super.num_inodes / (block_size/INODE_SIZE);
  table_length = inode.length_1;
  memcpy(itable,&inode,INODE_SIZE);
  alloc_inodes++;

  // Now we must create inodes to reference the inode and block bitmaps
  // Inode 1 = inode bitmap
  inode.type = SPECIAL;
  inode.subtype = INODE_BITMAP;
  inode.size = _num_inodes * block_size;
  inode.start_block_1 = super.inode_table + i;
  inode.length_1 = _num_inodes;
  inode_block = inode.start_block_1;
  inode_length = inode.length_1;
  memcpy(&itable[INODE_SIZE], &inode, INODE_SIZE);
  alloc_inodes++;

  // Inode 2 = block bitmap
  inode.type = SPECIAL;
  inode.subtype = BLOCK_BITMAP;
  inode.size = _num_blocks * block_size;
  inode.start_block_1 = super.inode_table + i + _num_inodes;
  inode.length_1 = _num_blocks;
  block_block = inode.start_block_1;
  block_length = inode.length_1;
  memcpy(&itable[INODE_SIZE*2], &inode, INODE_SIZE);
  alloc_inodes++;

  // Let's also create an inode for the root directory
  // Inode 3 = root directory
  inode.type = SPECIAL;
  inode.subtype = DIRECTORY;
  inode.size = 0;
  inode.permissions = 0xFFFF; // this will have to change soon
  inode.links_count = 0;
  inode.access_time = inode.creation_time = inode.modification_time = 0;
  inode.user_id = inode.group_id = 0;
  inode.flags = 0;
  inode.start_block_1 = super.inode_table + i + _num_inodes + _num_blocks;
  inode.length_1 = 1;
  root_block = inode.start_block_1;
  memcpy(&itable[INODE_SIZE*3], &inode, INODE_SIZE);
  alloc_inodes++;
  alloc_blocks++;

  // I suppose we should actually create the . and .. files as well.  Geez,
  // this format stuff is intense!  You can't forget anything :)
  cout << "Creating root directory\n";
  // The first file is "." (Current directory)
  dir.file_inode = 3;         // Heh, it's pointing it's self! :)
  dir.attrib_inode = 0;       // Remember, inode 0 is considered null
  dir.record_length = 16;     // gotta dword align it
  dir.name_length = 1;
  dir.character_encoding = 0; // regular ascii
  strcpy((char*)&dir.name, ".");
  memcpy(block_buffer, &dir, 16);

  // The next file is ".." (Parent directory)
  dir.file_inode = 3;         // Root dir doesn't have a .. so point to self
  dir.attrib_inode = 0;       // Remember, inode 0 is considered null
  dir.record_length = 16;     // gotta dword align it
  dir.name_length = 2;
  dir.character_encoding = 0; // ascii
  strcpy((char*)&dir.name, "..");
  memcpy(&block_buffer[16], &dir, 16);

  // Okay, now we simply flag all allocated inodes and blocks that need be.
  // Remember to reflect this information in the superblock as well!
  cout << "Flagging " << alloc_blocks << " blocks as allocated\n";
  super.free_blocks -= alloc_blocks;
  i = 0;
  while(alloc_blocks > 8) {    // fill in all full block bitmap bytes
    bbitmap[i] = 255;
    alloc_blocks -= 8;
    i++;
  }
  switch(alloc_blocks) {       // then fill in the sub byte area
    case 8: bbitmap[i] = 255; break;
    case 7: bbitmap[i] = 127; break;
    case 6: bbitmap[i] = 63;  break;
    case 5: bbitmap[i] = 31;  break;
    case 4: bbitmap[i] = 15;  break;
    case 3: bbitmap[i] = 7;   break;
    case 2: bbitmap[i] = 3;   break;
    case 1: bbitmap[i] = 1;   break;
  }

  cout << "Flagging " << alloc_inodes << " inodes as allocated\n";
  super.free_inodes -= alloc_inodes;
  i = 0;
  while(alloc_inodes > 8) {    // fill in all full inode bitmap bytes
    ibitmap[i] = 255;
    alloc_inodes -= 8;
    i++;
  }
  switch(alloc_inodes) {       // then fill in the sub byte area
    case 8: ibitmap[i] = 255; break;
    case 7: ibitmap[i] = 127; break;
    case 6: ibitmap[i] = 63;  break;
    case 5: ibitmap[i] = 31;  break;
    case 4: ibitmap[i] = 15;  break;
    case 3: ibitmap[i] = 7;   break;
    case 2: ibitmap[i] = 3;   break;
    case 1: ibitmap[i] = 1;   break;
  }

  // Okay, now we're finally done and must write everything to disk;  The
  // superblock, the inode table, the inode bitmap, and block bitmap, and
  // lastly, the root directory.  Pfew!
  cout << "Writting to disk...\n";

  // -- Write the superblock to head 0, cylinder 0, sector 2
  // We have to do this manually, because the super block isn't
  // necessary on a block boundary, as write_blocks() expects.
  cout << "  Superblock\n";
  biosdisk(0,drive,0,0,0,0,0);
  if(biosdisk(0x03,drive,0,0,3,2, &super)) {
    // Try again... it may just be a disk change that's causing the error
    biosdisk(0,drive,0,0,0,0,0);
    biosdisk(0x03,drive,0,0,3,2, &super);
  }

  // -- Write the inode table
  cout << "  Inode Table\n";
  write_blocks(itable, super.inode_table, table_length);

  // -- Write the inode bitmap
  cout << "  Inode Bitmap\n";
  write_blocks(ibitmap, inode_block, inode_length);

  // -- Write the block bitmap
  cout << "  Block Bitmap\n";
  write_blocks(bbitmap, block_block, block_length);

  // -- Write the root directory
  cout << "  Root Directory\n";
  write_blocks(block_buffer, root_block, 1);

  // And now we're done!
  return;
}

/* This function is very simple.  It's as low level as is necessary
 * in the Poly One file system.  I'd like to thank Josh McDonald for
 * releasing his real mode disk system.  It's quite good, and I used
 * it as a base for my own.  I can't wait 'till I get into real IDE
 * hard drive programming though!  No more BIOS, thank God!
 */
void PFS::write_blocks(char *buffer, int start, int total) {
  int current_cylinder,
      current_sector,
      current_head;
  int sectors_left,
      sectors_thistime;
  int error;

  // Convert our start and total block counts to sector counts
  start *= sectors_per_block;
  total *= sectors_per_block;

  // Calculate actual sector, head and cylinder values from our
  // logical sector "start" variable.
  current_sector   = start % sectors;
  current_head     = (start / sectors) % heads;
  current_cylinder = (start / sectors) / heads;

  // Here's an annoyance of the BIOS, the sector values start at one.
  current_sector++;

  // Now check if we're going to have to move any heads.  If so,
  // we'll have to split up into multiple writes
  if(current_sector+total <= sectors) {
    sectors_thistime = total;
    sectors_left = 0;
  } else {
    sectors_thistime = sectors - current_sector + 1;
    sectors_left = total - sectors_thistime;
  }

  // Okay, now just loop until we've written each sector
  while(sectors_left+sectors_thistime > 0) {
    // Reset the drive and attempt a write
    biosdisk(0,drive,0,0,0,0,0);
    error = biosdisk(0x03, drive, current_head, current_cylinder,
                     current_sector, sectors_thistime, buffer);

    while(error == 6) {
      // Try again... it may just be a disk change that's causing the error
      biosdisk(0,drive,0,0,0,0,0);
      error = biosdisk(0x03, drive, current_head, current_cylinder,
                       current_sector, sectors_thistime, buffer);
    }

    // Report any errors and exit if necessary
    if(error) {
      cout << "Error " << error << " writting to disk\n";
      exit(1);
    }

    // Adjust the buffer and sector values after the write
    buffer += (sectors_thistime*512);
    sectors_thistime = 0;

    // If we've got another run of sectors to write...
    if(sectors_left) {
      // Goto the next head
      current_sector = 1;
      current_head++;
      // And possibly the next cyclinder
      if(current_head > heads) {
        current_head = 0;
        current_cylinder++;
      }
      // Lastly, check if we're going to need to have yet another
      // run after this one.
      if(sectors_left > sectors) {
        sectors_thistime = sectors;
        sectors_left -= sectors;
      } else {
        sectors_thistime = sectors_left;
        sectors_left = 0;
      }
    }
  }
  return;
}

#endif
